@@ -290,7 +290,7 @@ struct QV4ParticleData : public QV4::Object
 
 DEFINE_OBJECT_VTABLE(QV4ParticleData);
 
-class QV4ParticleDataDeletable : public QV8Engine::Deletable
+class QV4ParticleDataDeletable : public QV4::ExecutionEngine::Deletable
 {
 public:
     QV4ParticleDataDeletable(QV4::ExecutionEngine *engine);
@@ -518,9 +518,9 @@ QQuickV4ParticleData::QQuickV4ParticleData(QV4::ExecutionEngine* v4, QQuickParti
 
     QV4::Scope scope(v4);
     QV4ParticleDataDeletable *d = particleV4Data(scope.engine);
-    QV4::ScopedObject o(scope, v4->memoryManager->allocObject<QV4ParticleData>(datum, system));
+    QV4::ScopedObject o(scope, v4->memoryManager->allocate<QV4ParticleData>(datum, system));
     QV4::ScopedObject p(scope, d->proto.value());
-    o->setPrototype(p);
+    o->setPrototypeUnchecked(p);
     m_v4Value = o;
 }
 
@@ -528,9 +528,9 @@ QQuickV4ParticleData::~QQuickV4ParticleData()
 {
 }
 
-QQmlV4Handle QQuickV4ParticleData::v4Value() const
+QV4::ReturnedValue QQuickV4ParticleData::v4Value() const
 {
-    return QQmlV4Handle(m_v4Value.value());
+    return m_v4Value.value();
 }
 
 QT_END_NAMESPACE
