@@ -47,25 +47,7 @@
 
 using namespace QV4;
 
-QT_WARNING_SUPPRESS_GCC_TAUTOLOGICAL_COMPARE_ON
-
-const QV4::VTable QV4::ArrayData::static_vtbl = {
-    nullptr,
-    0,
-    0,
-    QV4::ArrayData::IsExecutionContext,
-    QV4::ArrayData::IsString,
-    QV4::ArrayData::IsObject,
-    QV4::ArrayData::IsFunctionObject,
-    QV4::ArrayData::IsErrorObject,
-    QV4::ArrayData::IsArrayData,
-    0,
-    QV4::ArrayData::MyType,
-    "ArrayData",
-    Q_VTABLE_FUNCTION(QV4::ArrayData, destroy),
-    ArrayData::Data::markObjects,
-    isEqualTo
-};
+DEFINE_MANAGED_VTABLE(ArrayData);
 
 const ArrayVTable SimpleArrayData::static_vtbl =
 {
@@ -99,18 +81,9 @@ const ArrayVTable SparseArrayData::static_vtbl =
     SparseArrayData::length
 };
 
-QT_WARNING_SUPPRESS_GCC_TAUTOLOGICAL_COMPARE_OFF
-
 Q_STATIC_ASSERT(sizeof(Heap::ArrayData) == sizeof(Heap::SimpleArrayData));
 Q_STATIC_ASSERT(sizeof(Heap::ArrayData) == sizeof(Heap::SparseArrayData));
 
-static Q_ALWAYS_INLINE void storeValue(ReturnedValue *target, uint value)
-{
-    Value v;
-    v.setEmpty(value);
-    *target = v.asReturnedValue();
-}
-
 void Heap::ArrayData::markObjects(Heap::Base *base, MarkStack *stack)
 {
     ArrayData *a = static_cast<ArrayData *>(base);
@@ -145,8 +118,6 @@ void ArrayData::realloc(Object *o, Type newType, uint requested, bool enforceAtt
         if (d->type() > newType)
             newType = d->type();
     }
-    if (enforceAttributes && newType == Heap::ArrayData::Simple)
-        newType = Heap::ArrayData::Complex;
 
     while (alloc < requested)
         alloc *= 2;
@@ -195,7 +166,7 @@ void ArrayData::realloc(Object *o, Type newType, uint requested, bool enforceAtt
 
     Heap::SparseArrayData *sparse = static_cast<Heap::SparseArrayData *>(newData->d());
 
-    ReturnedValue *lastFree;
+    Value *lastFree;
     if (d && d->type() == Heap::ArrayData::Sparse) {
         Heap::SparseArrayData *old = static_cast<Heap::SparseArrayData *>(d->d());
         sparse->sparse = old->sparse;
@@ -204,29 +175,29 @@ void ArrayData::realloc(Object *o, Type newType, uint requested, bool enforceAtt
     } else {
         sparse->sparse = new SparseArray;
         lastFree = &sparse->sparse->freeList;
-        storeValue(lastFree, 0);
+        *lastFree = Encode(0);
         for (uint i = 0; i < toCopy; ++i) {
             if (!sparse->values[i].isEmpty()) {
                 SparseArrayNode *n = sparse->sparse->insert(i);
                 n->value = i;
             } else {
-                storeValue(lastFree, i);
+                *lastFree = Encode(i);
                 sparse->values.values[i].setEmpty();
-                lastFree = &sparse->values.values[i].rawValueRef();
+                lastFree = &sparse->values.values[i];
             }
         }
     }
 
     if (toCopy < sparse->values.alloc) {
         for (uint i = toCopy; i < sparse->values.alloc; ++i) {
-            storeValue(lastFree, i);
+            *lastFree = Encode(i);
             sparse->values.values[i].setEmpty();
-            lastFree = &sparse->values.values[i].rawValueRef();
+            lastFree = &sparse->values.values[i];
         }
     }
-    storeValue(lastFree, UINT_MAX);
+    *lastFree = Encode(-1);
 
-    Q_ASSERT(Value::fromReturnedValue(sparse->sparse->freeList).isEmpty());
+    Q_ASSERT(sparse->sparse->freeList.isInteger());
     // ### Could explicitly free the old data
 }
 
@@ -248,7 +219,7 @@ ReturnedValue SimpleArrayData::get(const Heap::ArrayData *d, uint index)
 {
     const Heap::SimpleArrayData *dd = static_cast<const Heap::SimpleArrayData *>(d);
     if (index >= dd->values.size)
-        return Primitive::emptyValue().asReturnedValue();
+        return Value::emptyValue().asReturnedValue();
     return dd->data(index).asReturnedValue();
 }
 
@@ -273,7 +244,7 @@ bool SimpleArrayData::del(Object *o, uint index)
         return true;
 
     if (!dd->attrs || dd->attrs[index].isConfigurable()) {
-        dd->setData(o->engine(), index, Primitive::emptyValue());
+        dd->setData(o->engine(), index, Value::emptyValue());
         if (dd->attrs)
             dd->attrs[index] = Attr_Data;
         return true;
@@ -355,7 +326,7 @@ bool SimpleArrayData::putArray(Object *o, uint index, const Value *values, uint
     }
     QV4::ExecutionEngine *e = o->engine();
     for (uint i = dd->values.size; i < index; ++i)
-        dd->setData(e, i, Primitive::emptyValue());
+        dd->setData(e, i, Value::emptyValue());
     for (uint i = 0; i < n; ++i)
         dd->setData(e, index + i, values[i]);
     dd->values.size = qMax(dd->values.size, index + n);
@@ -368,12 +339,12 @@ void SparseArrayData::free(Heap::ArrayData *d, uint idx)
     Value *v = d->values.values + idx;
     if (d->attrs && d->attrs[idx].isAccessor()) {
         // double slot, free both. Order is important, so we have a double slot for allocation again afterwards.
-        v[1].setEmpty(Value::fromReturnedValue(d->sparse->freeList).emptyValue());
-        v[0].setEmpty(idx + 1);
+        v[1] = d->sparse->freeList;
+        v[0] = Encode(idx + 1);
     } else {
-        v->setEmpty(Value::fromReturnedValue(d->sparse->freeList).emptyValue());
+        *v = d->sparse->freeList;
     }
-    d->sparse->freeList = Primitive::emptyValue(idx).asReturnedValue();
+    d->sparse->freeList = Encode(idx);
     if (d->attrs)
         d->attrs[idx].clear();
 }
@@ -390,36 +361,34 @@ uint SparseArrayData::allocate(Object *o, bool doubleSlot)
     Q_ASSERT(o->d()->arrayData->type == Heap::ArrayData::Sparse);
     Heap::SimpleArrayData *dd = o->d()->arrayData.cast<Heap::SimpleArrayData>();
     if (doubleSlot) {
-        ReturnedValue *last = &dd->sparse->freeList;
+        Value *last = &dd->sparse->freeList;
         while (1) {
-            if (Value::fromReturnedValue(*last).value() == UINT_MAX) {
+            if (last->int_32() == -1) {
                 reallocate(o, dd->values.alloc + 2, true);
                 dd = o->d()->arrayData.cast<Heap::SimpleArrayData>();
                 last = &dd->sparse->freeList;
-                Q_ASSERT(Value::fromReturnedValue(*last).value() != UINT_MAX);
+                Q_ASSERT(last->int_32() != -1);
             }
 
-            Q_ASSERT(dd->values[Value::fromReturnedValue(*last).value()].value() != Value::fromReturnedValue(*last).value());
-            if (dd->values[Value::fromReturnedValue(*last).value()].value() == (Value::fromReturnedValue(*last).value() + 1)) {
+            Q_ASSERT(dd->values[static_cast<uint>(last->int_32())].int_32() != last->int_32());
+            if (dd->values[static_cast<uint>(last->int_32())].int_32() == last->int_32() + 1) {
                 // found two slots in a row
-                uint idx = Value::fromReturnedValue(*last).emptyValue();
-                Value lastV = Value::fromReturnedValue(*last);
-                lastV.setEmpty(dd->values[lastV.emptyValue() + 1].value());
-                *last = lastV.rawValue();
+                uint idx = static_cast<uint>(last->int_32());
+                *last = Encode(dd->values[static_cast<uint>(last->int_32()) + 1].int_32());
                 dd->attrs[idx] = Attr_Accessor;
                 return idx;
             }
-            last = &dd->values.values[Value::fromReturnedValue(*last).value()].rawValueRef();
+            last = &dd->values.values[last->int_32()];
         }
     } else {
-        if (Value::fromReturnedValue(dd->sparse->freeList).value() == UINT_MAX) {
+        if (dd->sparse->freeList.int_32() == -1) {
             reallocate(o, dd->values.alloc + 1, false);
             dd = o->d()->arrayData.cast<Heap::SimpleArrayData>();
         }
-        uint idx = Value::fromReturnedValue(dd->sparse->freeList).value();
-        Q_ASSERT(idx != UINT_MAX);
-        dd->sparse->freeList = dd->values[idx].asReturnedValue();
-        Q_ASSERT(Value::fromReturnedValue(dd->sparse->freeList).isEmpty());
+        Q_ASSERT(dd->sparse->freeList.int_32() != -1);
+        uint idx = static_cast<uint>(dd->sparse->freeList.int_32());
+        dd->sparse->freeList = dd->values[idx];
+        Q_ASSERT(dd->sparse->freeList.isInteger());
         if (dd->attrs)
             dd->attrs[idx] = Attr_Data;
         return idx;
@@ -431,7 +400,7 @@ ReturnedValue SparseArrayData::get(const Heap::ArrayData *d, uint index)
     const Heap::SparseArrayData *s = static_cast<const Heap::SparseArrayData *>(d);
     index = s->mappedIndex(index);
     if (index == UINT_MAX)
-        return Primitive::emptyValue().asReturnedValue();
+        return Value::emptyValue().asReturnedValue();
     return s->values[index].asReturnedValue();
 }
 
@@ -474,14 +443,14 @@ bool SparseArrayData::del(Object *o, uint index)
 
     if (isAccessor) {
         // free up both indices
-        dd->values.values[pidx + 1].setEmpty(Value::fromReturnedValue(dd->sparse->freeList).emptyValue());
-        dd->values.values[pidx].setEmpty(pidx + 1);
+        dd->values.values[pidx + 1] = dd->sparse->freeList;
+        dd->values.values[pidx] = Encode(pidx + 1);
     } else {
         Q_ASSERT(dd->type == Heap::ArrayData::Sparse);
-        dd->values.values[pidx].setEmpty(Value::fromReturnedValue(dd->sparse->freeList).emptyValue());
+        dd->values.values[pidx] = dd->sparse->freeList;
     }
 
-    dd->sparse->freeList = Primitive::emptyValue(pidx).asReturnedValue();
+    dd->sparse->freeList = Encode(pidx);
     dd->sparse->erase(n);
     return true;
 }
@@ -593,7 +562,7 @@ uint ArrayData::append(Object *obj, ArrayObject *otherObj, uint n)
     if (!other || ArgumentsObject::isNonStrictArgumentsObject(otherObj)) {
         ScopedValue v(scope);
         for (uint i = 0; i < n; ++i)
-            obj->arraySet(oldSize + i, (v = otherObj->getIndexed(i)));
+            obj->arraySet(oldSize + i, (v = otherObj->get(i)));
     } else if (other && other->isSparse()) {
         Heap::SparseArrayData *os = static_cast<Heap::SparseArrayData *>(other->d());
         if (other->hasAttributes()) {
@@ -617,7 +586,7 @@ uint ArrayData::append(Object *obj, ArrayObject *otherObj, uint n)
         obj->arrayPut(oldSize, os->values.data() + os->offset, chunk);
         toCopy -= chunk;
         if (toCopy)
-            obj->arrayPut(oldSize + chunk, os->values.data(), toCopy);
+            obj->setArrayLength(oldSize + chunk + toCopy);
     }
 
     return oldSize + n;
@@ -635,7 +604,7 @@ void ArrayData::insert(Object *o, uint index, const Value *v, bool isAccessor)
             if (index >= d->values.size) {
                 // mark possible hole in the array
                 for (uint i = d->values.size; i < index; ++i)
-                    d->setData(o->engine(), i, Primitive::emptyValue());
+                    d->setData(o->engine(), i, Value::emptyValue());
                 d->values.size = index + 1;
             }
             d->setData(o->engine(), index, *v);
@@ -685,11 +654,19 @@ bool ArrayElementLessThan::operator()(Value v1, Value v2) const
         jsCallData->args[0] = v1;
         jsCallData->args[1] = v2;
         result = o->call(jsCallData);
+        if (scope.hasException())
+            return false;
 
         return result->toNumber() < 0;
     }
     ScopedString p1s(scope, v1.toString(scope.engine));
     ScopedString p2s(scope, v2.toString(scope.engine));
+
+    if (!p1s)
+        return false;
+    if (!p2s)
+        return true;
+
     return p1s->toQString() < p2s->toQString();
 }
 
@@ -828,7 +805,7 @@ void ArrayData::sort(ExecutionEngine *engine, Object *thisObject, const Value &c
                         break;
                 Q_ASSERT(!d->attrs || !d->attrs[len].isAccessor());
                 d->setData(engine, i, d->data(len));
-                d->setData(engine, len, Primitive::emptyValue());
+                d->setData(engine, len, Value::emptyValue());
             }
         }
 
