@@ -44,8 +44,10 @@
 #include "qv4objectproto_p.h"
 #include <private/qv4mm_p.h>
 #include "qv4scopedvalue_p.h"
-#include "qv4alloca_p.h"
+#include "qv4symbol_p.h"
+#include <private/qv4alloca_p.h>
 #include "qv4jscall_p.h"
+#include "qv4stringiterator_p.h"
 #include <QtCore/QDateTime>
 #include <QtCore/QDebug>
 #include <QtCore/QStringList>
@@ -72,14 +74,14 @@ void Heap::StringObject::init()
     Object::init();
     Q_ASSERT(vtable() == QV4::StringObject::staticVTable());
     string.set(internalClass->engine, internalClass->engine->id_empty()->d());
-    setProperty(internalClass->engine, LengthPropertyIndex, Primitive::fromInt32(0));
+    setProperty(internalClass->engine, LengthPropertyIndex, Value::fromInt32(0));
 }
 
 void Heap::StringObject::init(const QV4::String *str)
 {
     Object::init();
     string.set(internalClass->engine, str->d());
-    setProperty(internalClass->engine, LengthPropertyIndex, Primitive::fromInt32(length()));
+    setProperty(internalClass->engine, LengthPropertyIndex, Value::fromInt32(length()));
 }
 
 Heap::String *Heap::StringObject::getIndex(uint index) const
@@ -95,45 +97,71 @@ uint Heap::StringObject::length() const
     return string->length();
 }
 
-bool StringObject::deleteIndexedProperty(Managed *m, uint index)
+bool StringObject::virtualDeleteProperty(Managed *m, PropertyKey id)
 {
-    ExecutionEngine *v4 = static_cast<StringObject *>(m)->engine();
-    Scope scope(v4);
-    Scoped<StringObject> o(scope, m->as<StringObject>());
-    Q_ASSERT(!!o);
-
-    if (index < static_cast<uint>(o->d()->string->toQString().length()))
-        return false;
-    return true;
+    Q_ASSERT(m->as<StringObject>());
+    if (id.isArrayIndex()) {
+        StringObject *o = static_cast<StringObject *>(m);
+        uint index = id.asArrayIndex();
+        if (index < static_cast<uint>(o->d()->string->toQString().length()))
+            return false;
+    }
+    return Object::virtualDeleteProperty(m, id);
 }
 
-void StringObject::advanceIterator(Managed *m, ObjectIterator *it, Value *name, uint *index, Property *p, PropertyAttributes *attrs)
+struct StringObjectOwnPropertyKeyIterator : ObjectOwnPropertyKeyIterator
 {
-    name->setM(nullptr);
-    StringObject *s = static_cast<StringObject *>(m);
+    ~StringObjectOwnPropertyKeyIterator() override = default;
+    PropertyKey next(const QV4::Object *o, Property *pd = nullptr, PropertyAttributes *attrs = nullptr) override;
+
+};
+
+PropertyKey StringObjectOwnPropertyKeyIterator::next(const QV4::Object *o, Property *pd, PropertyAttributes *attrs)
+{
+    const StringObject *s = static_cast<const StringObject *>(o);
     uint slen = s->d()->string->toQString().length();
-    if (it->arrayIndex <= slen) {
-        while (it->arrayIndex < slen) {
-            *index = it->arrayIndex;
-            ++it->arrayIndex;
-            PropertyAttributes a;
-            Property pd;
-            s->getOwnProperty(*index, &a, &pd);
-            if (!(it->flags & ObjectIterator::EnumerableOnly) || a.isEnumerable()) {
-                *attrs = a;
-                p->copy(&pd, a);
-                return;
-            }
-        }
+    if (arrayIndex < slen) {
+        uint index = arrayIndex;
+        ++arrayIndex;
+        if (attrs)
+            *attrs = Attr_NotConfigurable|Attr_NotWritable;
+        if (pd)
+            pd->value = s->getIndex(index);
+        return PropertyKey::fromArrayIndex(index);
+    } else if (arrayIndex == slen) {
         if (s->arrayData()) {
-            it->arrayNode = s->sparseBegin();
+            SparseArrayNode *arrayNode = s->sparseBegin();
             // iterate until we're past the end of the string
-            while (it->arrayNode && it->arrayNode->key() < slen)
-                it->arrayNode = it->arrayNode->nextNode();
+            while (arrayNode && arrayNode->key() < slen)
+                arrayNode = arrayNode->nextNode();
         }
     }
 
-    return Object::advanceIterator(m, it, name, index, p, attrs);
+    return ObjectOwnPropertyKeyIterator::next(o, pd, attrs);
+}
+
+OwnPropertyKeyIterator *StringObject::virtualOwnPropertyKeys(const Object *m, Value *target)
+{
+    *target = *m;
+    return new StringObjectOwnPropertyKeyIterator;
+}
+
+PropertyAttributes StringObject::virtualGetOwnProperty(const Managed *m, PropertyKey id, Property *p)
+{
+    PropertyAttributes attributes = Object::virtualGetOwnProperty(m, id, p);
+    if (attributes != Attr_Invalid)
+        return attributes;
+
+    if (id.isArrayIndex()) {
+        const uint index = id.asArrayIndex();
+        const auto s = static_cast<const StringObject *>(m);
+        if (index < uint(s->d()->string->toQString().length())) {
+            if (p)
+                p->value = s->getIndex(index);
+            return Attr_NotConfigurable|Attr_NotWritable;
+        }
+    }
+    return Object::virtualGetOwnProperty(m, id, p);
 }
 
 DEFINE_OBJECT_VTABLE(StringCtor);
@@ -143,7 +171,7 @@ void Heap::StringCtor::init(QV4::ExecutionContext *scope)
     Heap::FunctionObject::init(scope, QStringLiteral("String"));
 }
 
-ReturnedValue StringCtor::callAsConstructor(const FunctionObject *f, const Value *argv, int argc)
+ReturnedValue StringCtor::virtualCallAsConstructor(const FunctionObject *f, const Value *argv, int argc, const Value *newTarget)
 {
     ExecutionEngine *v4 = static_cast<const Object *>(f)->engine();
     Scope scope(v4);
@@ -152,16 +180,105 @@ ReturnedValue StringCtor::callAsConstructor(const FunctionObject *f, const Value
         value = argv[0].toString(v4);
     else
         value = v4->newString();
-    return Encode(v4->newStringObject(value));
+    CHECK_EXCEPTION();
+    ReturnedValue o = Encode(v4->newStringObject(value));
+
+    if (!newTarget)
+        return o;
+    ScopedObject obj(scope, o);
+    obj->setProtoFromNewTarget(newTarget);
+    return obj->asReturnedValue();
 }
 
-ReturnedValue StringCtor::call(const FunctionObject *m, const Value *, const Value *argv, int argc)
+ReturnedValue StringCtor::virtualCall(const FunctionObject *m, const Value *, const Value *argv, int argc)
 {
     ExecutionEngine *v4 = m->engine();
-    if (argc)
-        return argv[0].toString(v4)->asReturnedValue();
-    else
+    if (!argc)
         return v4->newString()->asReturnedValue();
+    if (argv[0].isSymbol())
+        return v4->newString(argv[0].symbolValue()->descriptiveString())->asReturnedValue();
+    return argv[0].toString(v4)->asReturnedValue();
+}
+
+ReturnedValue StringCtor::method_fromCharCode(const FunctionObject *b, const Value *, const Value *argv, int argc)
+{
+    QString str(argc, Qt::Uninitialized);
+    QChar *ch = str.data();
+    for (int i = 0, ei = argc; i < ei; ++i) {
+        *ch = QChar(argv[i].toUInt16());
+        ++ch;
+    }
+    *ch = 0;
+    return Encode(b->engine()->newString(str));
+}
+
+
+
+ReturnedValue StringCtor::method_fromCodePoint(const FunctionObject *f, const Value *, const Value *argv, int argc)
+{
+    ExecutionEngine *e = f->engine();
+    QString result(argc*2, Qt::Uninitialized); // assume worst case
+    QChar *ch = result.data();
+    for (int i = 0; i < argc; ++i) {
+        double num = argv[i].toNumber();
+        if (e->hasException)
+            return Encode::undefined();
+        int cp = static_cast<int>(num);
+        if (cp != num || cp < 0 || cp > 0x10ffff)
+            return e->throwRangeError(QStringLiteral("String.fromCodePoint: argument out of range."));
+        if (cp > 0xffff) {
+            *ch = QChar::highSurrogate(cp);
+            ++ch;
+            *ch = QChar::lowSurrogate(cp);
+        } else {
+            *ch = cp;
+        }
+        ++ch;
+    }
+    *ch = 0;
+    result.truncate(ch - result.constData());
+    return e->newString(result)->asReturnedValue();
+}
+
+ReturnedValue StringCtor::method_raw(const FunctionObject *f, const Value *, const Value *argv, int argc)
+{
+    Scope scope(f);
+    if (!argc)
+        return scope.engine->throwTypeError();
+
+    ScopedObject cooked(scope, argv[0].toObject(scope.engine));
+    if (!cooked)
+        return scope.engine->throwTypeError();
+    ScopedString rawString(scope, scope.engine->newIdentifier(QStringLiteral("raw")));
+    ScopedValue rawValue(scope, cooked->get(rawString));
+    ScopedObject raw(scope, rawValue->toObject(scope.engine));
+    if (scope.hasException())
+        return Encode::undefined();
+
+    ++argv;
+    --argc;
+
+    QString result;
+    uint literalSegments = raw->getLength();
+    if (!literalSegments)
+        return scope.engine->id_empty()->asReturnedValue();
+
+    uint nextIndex = 0;
+    ScopedValue val(scope);
+    while (1) {
+        val = raw->get(nextIndex);
+        result += val->toQString();
+        if (scope.engine->hasException)
+            return Encode::undefined();
+        if (nextIndex + 1 == literalSegments)
+            return scope.engine->newString(result)->asReturnedValue();
+
+        if (nextIndex < static_cast<uint>(argc))
+            result += argv[nextIndex].toQString();
+        if (scope.engine->hasException)
+            return Encode::undefined();
+        ++nextIndex;
+    }
 }
 
 void StringPrototype::init(ExecutionEngine *engine, Object *ctor)
@@ -169,15 +286,24 @@ void StringPrototype::init(ExecutionEngine *engine, Object *ctor)
     Scope scope(engine);
     ScopedObject o(scope);
 
+    // need to set this once again, as these were not fully defined when creating the string proto
+    Heap::InternalClass *ic = scope.engine->classes[ExecutionEngine::Class_StringObject]->changePrototype(scope.engine->objectPrototype()->d());
+    d()->internalClass.set(scope.engine, ic);
+    d()->string.set(scope.engine, scope.engine->id_empty()->d());
+    setProperty(scope.engine, Heap::StringObject::LengthPropertyIndex, Value::fromInt32(0));
+
     ctor->defineReadonlyProperty(engine->id_prototype(), (o = this));
-    ctor->defineReadonlyProperty(engine->id_length(), Primitive::fromInt32(1));
-    ctor->defineDefaultProperty(QStringLiteral("fromCharCode"), method_fromCharCode, 1);
+    ctor->defineReadonlyConfigurableProperty(engine->id_length(), Value::fromInt32(1));
+    ctor->defineDefaultProperty(QStringLiteral("fromCharCode"), StringCtor::method_fromCharCode, 1);
+    ctor->defineDefaultProperty(QStringLiteral("fromCodePoint"), StringCtor::method_fromCodePoint, 1);
+    ctor->defineDefaultProperty(QStringLiteral("raw"), StringCtor::method_raw, 1);
 
     defineDefaultProperty(QStringLiteral("constructor"), (o = ctor));
     defineDefaultProperty(engine->id_toString(), method_toString);
     defineDefaultProperty(engine->id_valueOf(), method_toString); // valueOf and toString are identical
     defineDefaultProperty(QStringLiteral("charAt"), method_charAt, 1);
     defineDefaultProperty(QStringLiteral("charCodeAt"), method_charCodeAt, 1);
+    defineDefaultProperty(QStringLiteral("codePointAt"), method_codePointAt, 1);
     defineDefaultProperty(QStringLiteral("concat"), method_concat, 1);
     defineDefaultProperty(QStringLiteral("endsWith"), method_endsWith, 1);
     defineDefaultProperty(QStringLiteral("indexOf"), method_indexOf, 1);
@@ -185,6 +311,9 @@ void StringPrototype::init(ExecutionEngine *engine, Object *ctor)
     defineDefaultProperty(QStringLiteral("lastIndexOf"), method_lastIndexOf, 1);
     defineDefaultProperty(QStringLiteral("localeCompare"), method_localeCompare, 1);
     defineDefaultProperty(QStringLiteral("match"), method_match, 1);
+    defineDefaultProperty(QStringLiteral("normalize"), method_normalize, 0);
+    defineDefaultProperty(QStringLiteral("padEnd"), method_padEnd, 1);
+    defineDefaultProperty(QStringLiteral("padStart"), method_padStart, 1);
     defineDefaultProperty(QStringLiteral("repeat"), method_repeat, 1);
     defineDefaultProperty(QStringLiteral("replace"), method_replace, 2);
     defineDefaultProperty(QStringLiteral("search"), method_search, 1);
@@ -198,6 +327,7 @@ void StringPrototype::init(ExecutionEngine *engine, Object *ctor)
     defineDefaultProperty(QStringLiteral("toUpperCase"), method_toUpperCase);
     defineDefaultProperty(QStringLiteral("toLocaleUpperCase"), method_toLocaleUpperCase);
     defineDefaultProperty(QStringLiteral("trim"), method_trim);
+    defineDefaultProperty(engine->symbol_iterator(), method_iterator);
 }
 
 static Heap::String *thisAsString(ExecutionEngine *v4, const QV4::Value *thisObject)
@@ -270,6 +400,29 @@ ReturnedValue StringPrototype::method_charCodeAt(const FunctionObject *b, const
     return Encode(qt_qnan());
 }
 
+ReturnedValue StringPrototype::method_codePointAt(const FunctionObject *f, const Value *thisObject, const Value *argv, int argc)
+{
+    ExecutionEngine *v4 = f->engine();
+    QString value = getThisString(v4, thisObject);
+    if (v4->hasException)
+        return QV4::Encode::undefined();
+
+    int index = argc ? argv[0].toInteger() : 0;
+    if (v4->hasException)
+        return QV4::Encode::undefined();
+
+    if (index < 0 || index >= value.size())
+        return Encode::undefined();
+
+    uint first = value.at(index).unicode();
+    if (QChar::isHighSurrogate(first) && index + 1 < value.size()) {
+        uint second = value.at(index + 1).unicode();
+        if (QChar::isLowSurrogate(second))
+            return Encode(QChar::surrogateToUcs4(first, second));
+    }
+    return Encode(first);
+}
+
 ReturnedValue StringPrototype::method_concat(const FunctionObject *b, const Value *thisObject, const Value *argv, int argc)
 {
     ExecutionEngine *v4 = b->engine();
@@ -298,12 +451,11 @@ ReturnedValue StringPrototype::method_endsWith(const FunctionObject *b, const Va
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    QString searchString;
-    if (argc) {
-        if (argv[0].as<RegExpObject>())
-            return v4->throwTypeError();
-        searchString = argv[0].toQString();
-    }
+    if (argc && argv[0].as<RegExpObject>())
+        return v4->throwTypeError();
+    QString searchString = (argc ? argv[0] : Value::undefinedValue()).toQString();
+    if (v4->hasException)
+        return Encode::undefined();
 
     int pos = value.length();
     if (argc > 1)
@@ -323,9 +475,9 @@ ReturnedValue StringPrototype::method_indexOf(const FunctionObject *b, const Val
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    QString searchString;
-    if (argc)
-        searchString = argv[0].toQString();
+    QString searchString = (argc ? argv[0] : Value::undefinedValue()).toQString();
+    if (v4->hasException)
+        return Encode::undefined();
 
     int pos = 0;
     if (argc > 1)
@@ -345,12 +497,11 @@ ReturnedValue StringPrototype::method_includes(const FunctionObject *b, const Va
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    QString searchString;
-    if (argc) {
-        if (argv[0].as<RegExpObject>())
-            return v4->throwTypeError();
-        searchString = argv[0].toQString();
-    }
+    if (argc && argv[0].as<RegExpObject>())
+        return v4->throwTypeError();
+    QString searchString = (argc ? argv[0] : Value::undefinedValue()).toQString();
+    if (v4->hasException)
+        return Encode::undefined();
 
     int pos = 0;
     if (argc > 1) {
@@ -374,9 +525,9 @@ ReturnedValue StringPrototype::method_lastIndexOf(const FunctionObject *b, const
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    QString searchString;
-    if (argc)
-        searchString = argv[0].toQString();
+    QString searchString = (argc ? argv[0] : Value::undefinedValue()).toQString();
+    if (v4->hasException)
+        return Encode::undefined();
 
     double position = argc > 1 ? RuntimeHelpers::toNumber(argv[1]) : +qInf();
     if (std::isnan(position))
@@ -400,7 +551,7 @@ ReturnedValue StringPrototype::method_localeCompare(const FunctionObject *b, con
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    const QString that = (argc ? argv[0] : Primitive::undefinedValue()).toQString();
+    const QString that = (argc ? argv[0] : Value::undefinedValue()).toQString();
     return Encode(QString::localeAwareCompare(value, that));
 }
 
@@ -411,50 +562,147 @@ ReturnedValue StringPrototype::method_match(const FunctionObject *b, const Value
         return v4->throwTypeError();
 
     Scope scope(v4);
+    if (argc && !argv[0].isNullOrUndefined()) {
+        ScopedObject r(scope, argv[0].toObject(scope.engine));
+        if (scope.hasException())
+            return Encode::undefined();
+        ScopedValue f(scope, r->get(scope.engine->symbol_match()));
+        if (!f->isNullOrUndefined()) {
+            ScopedFunctionObject fo(scope, f);
+            if (!fo)
+                return scope.engine->throwTypeError();
+            return checkedResult(scope.engine, fo->call(r, thisObject, 1));
+        }
+    }
+
     ScopedString s(scope, thisObject->toString(v4));
     if (v4->hasException)
         return Encode::undefined();
 
-    Scoped<RegExpObject> that(scope, argc ? argv[0] : Primitive::undefinedValue());
+    Scoped<RegExpObject> that(scope, argc ? argv[0] : Value::undefinedValue());
     if (!that) {
         // convert args[0] to a regexp
-        that = RegExpCtor::callAsConstructor(b, argv, argc);
+        that = RegExpCtor::virtualCallAsConstructor(b, argv, argc, b);
         if (v4->hasException)
             return Encode::undefined();
     }
     Q_ASSERT(!!that);
 
-    bool global = that->global();
-
-    if (!global)
-        return RegExpPrototype::method_exec(b, that, s, 1);
-
-    // rx is now in thisObject
-    that->setLastIndex(0);
-    ScopedArrayObject a(scope, scope.engine->newArrayObject());
-
-    int previousLastIndex = 0;
-    uint n = 0;
-    while (1) {
-        Value result = Primitive::fromReturnedValue(RegExpPrototype::execFirstMatch(b, that, s, 1));
-        if (result.isNull())
-            break;
-        int index = that->lastIndex();
-        if (previousLastIndex == index) {
-            previousLastIndex = index + 1;
-            that->setLastIndex(previousLastIndex);
-        } else {
-            previousLastIndex = index;
-        }
-        a->arraySet(n, result);
-        ++n;
-    }
-    if (!n)
-        return Encode::null();
-    else
-        return a.asReturnedValue();
+    ScopedFunctionObject match(scope, that->get(scope.engine->symbol_match()));
+    if (!match)
+        return scope.engine->throwTypeError();
+    return checkedResult(scope.engine, match->call(that, s, 1));
 }
 
+ReturnedValue StringPrototype::method_normalize(const FunctionObject *f, const Value *thisObject, const Value *argv, int argc)
+{
+    ExecutionEngine *v4 = f->engine();
+    const QString value = getThisString(v4, thisObject);
+    if (v4->hasException)
+        return Encode::undefined();
+
+    QString::NormalizationForm form = QString::NormalizationForm_C;
+    if (argc >= 1 && !argv[0].isUndefined()) {
+        QString f = argv[0].toQString();
+        if (v4->hasException)
+            return Encode::undefined();
+        if (f == QLatin1String("NFC"))
+            form = QString::NormalizationForm_C;
+        else if (f == QLatin1String("NFD"))
+            form = QString::NormalizationForm_D;
+        else if (f == QLatin1String("NFKC"))
+            form = QString::NormalizationForm_KC;
+        else if (f == QLatin1String("NFKD"))
+            form = QString::NormalizationForm_KD;
+        else
+            return v4->throwRangeError(QLatin1String("String.prototype.normalize: Invalid normalization form."));
+    }
+    QString normalized = value.normalized(form);
+    return v4->newString(normalized)->asReturnedValue();
+}
+
+ReturnedValue StringPrototype::method_padEnd(const FunctionObject *f, const Value *thisObject, const Value *argv, int argc)
+{
+    ExecutionEngine *v4 = f->engine();
+    if (thisObject->isNullOrUndefined())
+        return v4->throwTypeError();
+
+    Scope scope(v4);
+    ScopedString s(scope, thisAsString(v4, thisObject));
+    if (v4->hasException)
+        return Encode::undefined();
+    if (!argc)
+        return s->asReturnedValue();
+
+    int maxLen = argv[0].toInteger();
+    if (maxLen <= s->d()->length())
+        return s->asReturnedValue();
+    QString fillString = (argc > 1 && !argv[1].isUndefined()) ? argv[1].toQString() : QString::fromLatin1(" ");
+    if (v4->hasException)
+        return Encode::undefined();
+
+    if (fillString.isEmpty())
+        return s->asReturnedValue();
+
+    QString padded = s->toQString();
+    int oldLength = padded.length();
+    int toFill = maxLen - oldLength;
+    padded.resize(maxLen);
+    QChar *ch = padded.data() + oldLength;
+    while (toFill) {
+        int copy = qMin(fillString.length(), toFill);
+        memcpy(ch, fillString.constData(), copy*sizeof(QChar));
+        toFill -= copy;
+        ch += copy;
+    }
+    *ch = 0;
+
+    return v4->newString(padded)->asReturnedValue();
+}
+
+ReturnedValue StringPrototype::method_padStart(const FunctionObject *f, const Value *thisObject, const Value *argv, int argc)
+{
+    ExecutionEngine *v4 = f->engine();
+    if (thisObject->isNullOrUndefined())
+        return v4->throwTypeError();
+
+    Scope scope(v4);
+    ScopedString s(scope, thisAsString(v4, thisObject));
+    if (v4->hasException)
+        return Encode::undefined();
+    if (!argc)
+        return s->asReturnedValue();
+
+    int maxLen = argv[0].toInteger();
+    if (maxLen <= s->d()->length())
+        return s->asReturnedValue();
+    QString fillString = (argc > 1 && !argv[1].isUndefined()) ? argv[1].toQString() : QString::fromLatin1(" ");
+    if (v4->hasException)
+        return Encode::undefined();
+
+    if (fillString.isEmpty())
+        return s->asReturnedValue();
+
+    QString original = s->toQString();
+    int oldLength = original.length();
+    int toFill = maxLen - oldLength;
+    QString padded;
+    padded.resize(maxLen);
+    QChar *ch = padded.data();
+    while (toFill) {
+        int copy = qMin(fillString.length(), toFill);
+        memcpy(ch, fillString.constData(), copy*sizeof(QChar));
+        toFill -= copy;
+        ch += copy;
+    }
+    memcpy(ch, original.constData(), oldLength*sizeof(QChar));
+    ch += oldLength;
+    *ch = 0;
+
+    return v4->newString(padded)->asReturnedValue();
+}
+
+
 ReturnedValue StringPrototype::method_repeat(const FunctionObject *b, const Value *thisObject, const Value *argv, int argc)
 {
     ExecutionEngine *v4 = b->engine();
@@ -462,7 +710,7 @@ ReturnedValue StringPrototype::method_repeat(const FunctionObject *b, const Valu
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    double repeats = (argc ? argv[0] : Primitive::undefinedValue()).toInteger();
+    double repeats = (argc ? argv[0] : Value::undefinedValue()).toInteger();
 
     if (repeats < 0 || qIsInf(repeats))
         return v4->throwRangeError(QLatin1String("Invalid count value"));
@@ -475,41 +723,51 @@ static void appendReplacementString(QString *result, const QString &input, const
     result->reserve(result->length() + replaceValue.length());
     for (int i = 0; i < replaceValue.length(); ++i) {
         if (replaceValue.at(i) == QLatin1Char('$') && i < replaceValue.length() - 1) {
-            ushort ch = replaceValue.at(++i).unicode();
+            ushort ch = replaceValue.at(i + 1).unicode();
             uint substStart = JSC::Yarr::offsetNoMatch;
             uint substEnd = JSC::Yarr::offsetNoMatch;
+            int skip = 0;
             if (ch == '$') {
                 *result += QChar(ch);
+                ++i;
                 continue;
             } else if (ch == '&') {
                 substStart = matchOffsets[0];
                 substEnd = matchOffsets[1];
+                skip = 1;
             } else if (ch == '`') {
                 substStart = 0;
                 substEnd = matchOffsets[0];
+                skip = 1;
             } else if (ch == '\'') {
                 substStart = matchOffsets[1];
                 substEnd = input.length();
-            } else if (ch >= '1' && ch <= '9') {
+                skip = 1;
+            } else if (ch >= '0' && ch <= '9') {
                 uint capture = ch - '0';
-                Q_ASSERT(capture > 0);
-                if (capture < static_cast<uint>(captureCount)) {
-                    substStart = matchOffsets[capture * 2];
-                    substEnd = matchOffsets[capture * 2 + 1];
-                }
-            } else if (ch == '0' && i < replaceValue.length() - 1) {
-                int capture = (ch - '0') * 10;
-                ch = replaceValue.at(++i).unicode();
-                if (ch >= '0' && ch <= '9') {
-                    capture += ch - '0';
-                    if (capture > 0 && capture < captureCount) {
-                        substStart = matchOffsets[capture * 2];
-                        substEnd = matchOffsets[capture * 2 + 1];
+                skip = 1;
+                if (i < replaceValue.length() - 2) {
+                    ch = replaceValue.at(i + 2).unicode();
+                    if (ch >= '0' && ch <= '9') {
+                        uint c = capture*10 + ch - '0';
+                        if (c < static_cast<uint>(captureCount)) {
+                            capture = c;
+                            skip = 2;
+                        }
                     }
                 }
+                if (capture > 0 && capture < static_cast<uint>(captureCount)) {
+                    substStart = matchOffsets[capture * 2];
+                    substEnd = matchOffsets[capture * 2 + 1];
+                } else {
+                    skip = 0;
+                }
             }
+            i += skip;
             if (substStart != JSC::Yarr::offsetNoMatch && substEnd != JSC::Yarr::offsetNoMatch)
                 *result += input.midRef(substStart, substEnd - substStart);
+            else if (skip == 0) // invalid capture reference. Taken as literal value
+                *result += replaceValue.at(i);
         } else {
             *result += replaceValue.at(i);
         }
@@ -532,7 +790,7 @@ ReturnedValue StringPrototype::method_replace(const FunctionObject *b, const Val
     uint *matchOffsets = _matchOffsets;
 
     Scope scope(b);
-    ScopedValue searchValue(scope, argc ? argv[0] : Primitive::undefinedValue());
+    ScopedValue searchValue(scope, argc ? argv[0] : Value::undefinedValue());
     Scoped<RegExpObject> regExp(scope, searchValue);
     if (regExp) {
         uint offset = 0;
@@ -559,8 +817,11 @@ ReturnedValue StringPrototype::method_replace(const FunctionObject *b, const Val
                 break;
             offset = qMax(offset + 1, matchOffsets[oldSize + 1]);
         }
-        if (regExp->global())
+        if (regExp->global()) {
             regExp->setLastIndex(0);
+            if (scope.hasException())
+                return Encode::undefined();
+        }
         numStringMatches = nMatchOffsets / (regExp->value()->captureCount() * 2);
         numCaptures = regExp->value()->captureCount();
     } else {
@@ -576,7 +837,7 @@ ReturnedValue StringPrototype::method_replace(const FunctionObject *b, const Val
 
     QString result;
     ScopedValue replacement(scope);
-    ScopedValue replaceValue(scope, argc > 1 ? argv[1] : Primitive::undefinedValue());
+    ScopedValue replaceValue(scope, argc > 1 ? argv[1] : Value::undefinedValue());
     ScopedFunctionObject searchCallback(scope, replaceValue);
     if (!!searchCallback) {
         result.reserve(string.length() + 10*numStringMatches);
@@ -588,7 +849,7 @@ ReturnedValue StringPrototype::method_replace(const FunctionObject *b, const Val
                 int idx = (i * numCaptures + k) * 2;
                 uint start = matchOffsets[idx];
                 uint end = matchOffsets[idx + 1];
-                entry = Primitive::undefinedValue();
+                entry = Value::undefinedValue();
                 if (start != JSC::Yarr::offsetNoMatch && end != JSC::Yarr::offsetNoMatch)
                     entry = scope.engine->newString(string.mid(start, end - start));
                 arguments[k] = entry;
@@ -596,11 +857,12 @@ ReturnedValue StringPrototype::method_replace(const FunctionObject *b, const Val
             uint matchStart = matchOffsets[i * numCaptures * 2];
             Q_ASSERT(matchStart >= static_cast<uint>(lastEnd));
             uint matchEnd = matchOffsets[i * numCaptures * 2 + 1];
-            arguments[numCaptures] = Primitive::fromUInt32(matchStart);
+            arguments[numCaptures] = Value::fromUInt32(matchStart);
             arguments[numCaptures + 1] = scope.engine->newString(string);
 
-            Value that = Primitive::undefinedValue();
+            Value that = Value::undefinedValue();
             replacement = searchCallback->call(&that, arguments, numCaptures + 2);
+            CHECK_EXCEPTION();
             result += string.midRef(lastEnd, matchStart - lastEnd);
             result += replacement->toQString();
             lastEnd = matchEnd;
@@ -638,7 +900,7 @@ ReturnedValue StringPrototype::method_search(const FunctionObject *b, const Valu
     if (scope.engine->hasException)
         return QV4::Encode::undefined();
 
-    Scoped<RegExpObject> regExp(scope, argc ? argv[0] : Primitive::undefinedValue());
+    Scoped<RegExpObject> regExp(scope, argc ? argv[0] : Value::undefinedValue());
     if (!regExp) {
         regExp = scope.engine->regExpCtor()->callAsConstructor(argv, 1);
         if (scope.engine->hasException)
@@ -695,8 +957,8 @@ ReturnedValue StringPrototype::method_split(const FunctionObject *b, const Value
         return QV4::Encode::undefined();
 
     Scope scope(v4);
-    ScopedValue separatorValue(scope, argc ? argv[0] : Primitive::undefinedValue());
-    ScopedValue limitValue(scope, argc > 1 ? argv[1] : Primitive::undefinedValue());
+    ScopedValue separatorValue(scope, argc ? argv[0] : Value::undefinedValue());
+    ScopedValue limitValue(scope, argc > 1 ? argv[1] : Value::undefinedValue());
 
     ScopedArrayObject array(scope, scope.engine->newArrayObject());
 
@@ -777,12 +1039,11 @@ ReturnedValue StringPrototype::method_startsWith(const FunctionObject *b, const
     if (v4->hasException)
         return QV4::Encode::undefined();
 
-    QString searchString;
-    if (argc) {
-        if (argv[0].as<RegExpObject>())
-            return v4->throwTypeError();
-        searchString = argv[0].toQString();
-    }
+    if (argc && argv[0].as<RegExpObject>())
+        return v4->throwTypeError();
+    QString searchString = (argc ? argv[0] : Value::undefinedValue()).toQString();
+    if (v4->hasException)
+        return Encode::undefined();
 
     int pos = 0;
     if (argc > 1)
@@ -816,8 +1077,8 @@ ReturnedValue StringPrototype::method_substr(const FunctionObject *b, const Valu
 
     length = qMin(qMax(length, 0.0), count - start);
 
-    qint32 x = Primitive::toInt32(start);
-    qint32 y = Primitive::toInt32(length);
+    qint32 x = Value::toInt32(start);
+    qint32 y = Value::toInt32(length);
     return Encode(v4->newString(value.mid(x, y)));
 }
 
@@ -892,17 +1153,6 @@ ReturnedValue StringPrototype::method_toLocaleUpperCase(const FunctionObject *b,
     return method_toUpperCase(b, thisObject, argv, argc);
 }
 
-ReturnedValue StringPrototype::method_fromCharCode(const FunctionObject *b, const Value *, const Value *argv, int argc)
-{
-    QString str(argc, Qt::Uninitialized);
-    QChar *ch = str.data();
-    for (int i = 0, ei = argc; i < ei; ++i) {
-        *ch = QChar(argv[i].toUInt16());
-        ++ch;
-    }
-    return Encode(b->engine()->newString(str));
-}
-
 ReturnedValue StringPrototype::method_trim(const FunctionObject *b, const Value *thisObject, const Value *, int)
 {
     ExecutionEngine *v4 = b->engine();
@@ -923,3 +1173,16 @@ ReturnedValue StringPrototype::method_trim(const FunctionObject *b, const Value
 
     return Encode(v4->newString(QString(chars + start, end - start + 1)));
 }
+
+
+
+ReturnedValue StringPrototype::method_iterator(const FunctionObject *b, const Value *thisObject, const Value *, int)
+{
+    Scope scope(b);
+    ScopedString s(scope, thisObject->toString(scope.engine));
+    if (!s || thisObject->isNullOrUndefined())
+        return scope.engine->throwTypeError();
+
+    Scoped<StringIteratorObject> si(scope, scope.engine->memoryManager->allocate<StringIteratorObject>(s->d(), scope.engine));
+    return si->asReturnedValue();
+}
