@@ -65,9 +65,7 @@
 
 #include "qqmlguard_p.h"
 #include "qqmlcontext_p.h"
-#include "qqmlpropertycache_p.h"
 
-#include <private/qv8engine_p.h>
 #include <private/qflagpointer_p.h>
 
 #include <private/qv4object_p.h>
@@ -146,7 +144,7 @@ class QQmlVMEMetaObjectEndpoint;
 class Q_QML_PRIVATE_EXPORT QQmlVMEMetaObject : public QQmlInterceptorMetaObject
 {
 public:
-    QQmlVMEMetaObject(QV4::ExecutionEngine *engine, QObject *obj, const QQmlRefPointer<QQmlPropertyCache> &cache, const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &qmlCompilationUnit, int qmlObjectId);
+    QQmlVMEMetaObject(QV4::ExecutionEngine *engine, QObject *obj, const QQmlRefPointer<QQmlPropertyCache> &cache, const QQmlRefPointer<QV4::ExecutableCompilationUnit> &qmlCompilationUnit, int qmlObjectId);
     ~QQmlVMEMetaObject() override;
 
     bool aliasTarget(int index, QObject **target, int *coreIndex, int *valueTypeIndex) const;
@@ -234,7 +232,7 @@ public:
 
     // keep a reference to the compilation unit in order to still
     // do property access when the context has been invalidated.
-    QQmlRefPointer<QV4::CompiledData::CompilationUnit> compilationUnit;
+    QQmlRefPointer<QV4::ExecutableCompilationUnit> compilationUnit;
     const QV4::CompiledData::Object *compiledObject;
 };
 
