@@ -45,6 +45,7 @@
 #include <private/qqmlglobal_p.h>
 
 #include <private/qqmlcomponent_p.h>
+#include <private/qqmlincubator_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -54,8 +55,8 @@ static const QQuickItemPrivate::ChangeTypes watchedChanges
     = QQuickItemPrivate::Geometry | QQuickItemPrivate::ImplicitWidth | QQuickItemPrivate::ImplicitHeight;
 
 QQuickLoaderPrivate::QQuickLoaderPrivate()
-    : item(nullptr), object(nullptr), component(nullptr), itemContext(nullptr), incubator(nullptr), updatingSize(false),
-      active(true), loadingFromSource(false), asynchronous(false)
+    : item(nullptr), object(nullptr), itemContext(nullptr), incubator(nullptr), updatingSize(false),
+      active(true), loadingFromSource(false), asynchronous(false), status(computeStatus())
 {
 }
 
@@ -111,11 +112,10 @@ void QQuickLoaderPrivate::clear()
         QObject::disconnect(component, SIGNAL(progressChanged(qreal)),
                 q, SIGNAL(progressChanged()));
         component->deleteLater();
-        component = nullptr;
+        component.setObject(nullptr, q);
     } else if (component) {
-        component = nullptr;
+        component.setObject(nullptr, q);
     }
-    componentStrongReference.clear();
     source = QUrl();
 
     if (item) {
@@ -379,7 +379,7 @@ void QQuickLoader::setActive(bool newVal)
             d->object = nullptr;
             emit itemChanged();
         }
-        emit statusChanged();
+        d->updateStatus();
     }
     emit activeChanged();
 }
@@ -432,7 +432,7 @@ void QQuickLoader::loadFromSource()
     Q_D(QQuickLoader);
     if (d->source.isEmpty()) {
         emit sourceChanged();
-        emit statusChanged();
+        d->updateStatus();
         emit progressChanged();
         emit itemChanged();
         return;
@@ -441,7 +441,7 @@ void QQuickLoader::loadFromSource()
     if (isComponentComplete()) {
         QQmlComponent::CompilationMode mode = d->asynchronous ? QQmlComponent::Asynchronous : QQmlComponent::PreferSynchronous;
         if (!d->component)
-            d->component = new QQmlComponent(qmlEngine(this), d->source, mode, this);
+            d->component.setObject(new QQmlComponent(qmlEngine(this), d->source, mode, this), this);
         d->load();
     }
 }
@@ -484,11 +484,7 @@ void QQuickLoader::setSourceComponent(QQmlComponent *comp)
 
     d->clear();
 
-    d->component = comp;
-    if (comp) {
-        if (QQmlData *ddata = QQmlData::get(comp))
-            d->componentStrongReference = ddata->jsWrapper;
-    }
+    d->component.setObject(comp, this);
     d->loadingFromSource = false;
 
     if (d->active)
@@ -507,7 +503,7 @@ void QQuickLoader::loadFromSourceComponent()
     Q_D(QQuickLoader);
     if (!d->component) {
         emit sourceComponentChanged();
-        emit statusChanged();
+        d->updateStatus();
         emit progressChanged();
         emit itemChanged();
         return;
@@ -593,10 +589,17 @@ void QQuickLoader::setSource(QQmlV4Function *args)
     if (ipvError)
         return;
 
+    // 1. If setSource is called with a valid url, clear the old component and its corresponding url
+    // 2. If setSource is called with an invalid url(e.g. empty url), clear the old component but
+    // hold the url for old one.(we will compare it with new url later and may update status of loader to Loader.Null)
+    QUrl oldUrl = d->source;
     d->clear();
     QUrl sourceUrl = d->resolveSourceUrl(args);
+    if (!sourceUrl.isValid())
+        d->source = oldUrl;
+
+    d->disposeInitialPropertyValues();
     if (!ipv->isUndefined()) {
-        d->disposeInitialPropertyValues();
         d->initialPropertyValues.set(args->v4engine(), ipv);
     }
     d->qmlCallingContext.set(scope.engine, scope.engine->qmlContext());
@@ -606,6 +609,7 @@ void QQuickLoader::setSource(QQmlV4Function *args)
 
 void QQuickLoaderPrivate::disposeInitialPropertyValues()
 {
+    initialPropertyValues.clear();
 }
 
 void QQuickLoaderPrivate::load()
@@ -622,7 +626,7 @@ void QQuickLoaderPrivate::load()
                 q, SLOT(_q_sourceLoaded()));
         QObject::connect(component, SIGNAL(progressChanged(qreal)),
                 q, SIGNAL(progressChanged()));
-        emit q->statusChanged();
+        updateStatus();
         emit q->progressChanged();
         if (loadingFromSource)
             emit q->sourceChanged();
@@ -669,7 +673,8 @@ void QQuickLoaderPrivate::setInitialState(QObject *obj)
     QV4::Scope scope(v4);
     QV4::ScopedValue ipv(scope, initialPropertyValues.value());
     QV4::Scoped<QV4::QmlContext> qmlContext(scope, qmlCallingContext.value());
-    d->initializeObjectWithInitialProperties(qmlContext, ipv, obj);
+    auto incubatorPriv = QQmlIncubatorPrivate::get(incubator);
+    d->initializeObjectWithInitialProperties(qmlContext, ipv, obj, incubatorPriv->requiredProperties());
 }
 
 void QQuickLoaderIncubator::statusChanged(Status status)
@@ -709,7 +714,7 @@ void QQuickLoaderPrivate::incubatorStateChanged(QQmlIncubator::Status status)
         emit q->sourceChanged();
     else
         emit q->sourceComponentChanged();
-    emit q->statusChanged();
+    updateStatus();
     emit q->progressChanged();
     if (status == QQmlIncubator::Ready)
         emit q->loaded();
@@ -726,7 +731,7 @@ void QQuickLoaderPrivate::_q_sourceLoaded()
             emit q->sourceChanged();
         else
             emit q->sourceComponentChanged();
-        emit q->statusChanged();
+        updateStatus();
         emit q->progressChanged();
         emit q->itemChanged(); //Like clearing source, emit itemChanged even if previous item was also null
         disposeInitialPropertyValues(); // cleanup
@@ -744,7 +749,7 @@ void QQuickLoaderPrivate::_q_sourceLoaded()
     component->create(*incubator, itemContext);
 
     if (incubator && incubator->status() == QQmlIncubator::Loading)
-        emit q->statusChanged();
+        updateStatus();
 }
 
 /*!
@@ -791,37 +796,7 @@ QQuickLoader::Status QQuickLoader::status() const
 {
     Q_D(const QQuickLoader);
 
-    if (!d->active)
-        return Null;
-
-    if (d->component) {
-        switch (d->component->status()) {
-        case QQmlComponent::Loading:
-            return Loading;
-        case QQmlComponent::Error:
-            return Error;
-        case QQmlComponent::Null:
-            return Null;
-        default:
-            break;
-        }
-    }
-
-    if (d->incubator) {
-        switch (d->incubator->status()) {
-        case QQmlIncubator::Loading:
-            return Loading;
-        case QQmlIncubator::Error:
-            return Error;
-        default:
-            break;
-        }
-    }
-
-    if (d->object)
-        return Ready;
-
-    return d->source.isEmpty() ? Null : Error;
+    return static_cast<Status>(d->status);
 }
 
 void QQuickLoader::componentComplete()
@@ -832,7 +807,7 @@ void QQuickLoader::componentComplete()
         if (d->loadingFromSource) {
             QQmlComponent::CompilationMode mode = d->asynchronous ? QQmlComponent::Asynchronous : QQmlComponent::PreferSynchronous;
             if (!d->component)
-                d->component = new QQmlComponent(qmlEngine(this), d->source, mode, this);
+                d->component.setObject(new QQmlComponent(qmlEngine(this), d->source, mode, this), this);
         }
         d->load();
     }
@@ -855,8 +830,6 @@ void QQuickLoader::itemChange(QQuickItem::ItemChange change, const QQuickItem::I
 
     This signal is emitted when the \l status becomes \c Loader.Ready, or on successful
     initial load.
-
-    The corresponding handler is \c onLoaded.
 */
 
 
@@ -995,6 +968,8 @@ QUrl QQuickLoaderPrivate::resolveSourceUrl(QQmlV4Function *args)
 {
     QV4::Scope scope(args->v4engine());
     QV4::ScopedValue v(scope, (*args)[0]);
+    if (v->isUndefined())
+        return QUrl();
     QString arg = v->toQString();
     if (arg.isEmpty())
         return QUrl();
@@ -1007,7 +982,7 @@ QUrl QQuickLoaderPrivate::resolveSourceUrl(QQmlV4Function *args)
 QV4::ReturnedValue QQuickLoaderPrivate::extractInitialPropertyValues(QQmlV4Function *args, QObject *loader, bool *error)
 {
     QV4::Scope scope(args->v4engine());
-    QV4::ScopedValue valuemap(scope, QV4::Primitive::undefinedValue());
+    QV4::ScopedValue valuemap(scope, QV4::Encode::undefined());
     if (args->length() >= 2) {
         QV4::ScopedValue v(scope, (*args)[1]);
         if (!v->isObject() || v->as<QV4::ArrayObject>()) {
@@ -1022,6 +997,51 @@ QV4::ReturnedValue QQuickLoaderPrivate::extractInitialPropertyValues(QQmlV4Funct
     return valuemap->asReturnedValue();
 }
 
+QQuickLoader::Status QQuickLoaderPrivate::computeStatus() const
+{
+    if (!active)
+        return QQuickLoader::Status::Null;
+
+    if (component) {
+        switch (component->status()) {
+        case QQmlComponent::Loading:
+            return QQuickLoader::Status::Loading;
+        case QQmlComponent::Error:
+            return QQuickLoader::Status::Error;
+        case QQmlComponent::Null:
+            return QQuickLoader::Status::Null;
+        default:
+            break;
+        }
+    }
+
+    if (incubator) {
+        switch (incubator->status()) {
+        case QQmlIncubator::Loading:
+            return QQuickLoader::Status::Loading;
+        case QQmlIncubator::Error:
+            return QQuickLoader::Status::Error;
+        default:
+            break;
+        }
+    }
+
+    if (object)
+        return QQuickLoader::Status::Ready;
+
+    return source.isEmpty() ? QQuickLoader::Status::Null : QQuickLoader::Status::Error;
+}
+
+void QQuickLoaderPrivate::updateStatus()
+{
+    Q_Q(QQuickLoader);
+    auto newStatus = computeStatus();
+    if (status != newStatus) {
+        status = newStatus;
+        emit q->statusChanged();
+    }
+}
+
 #include <moc_qquickloader_p.cpp>
 
 QT_END_NAMESPACE
