@@ -57,6 +57,10 @@
 
 QT_BEGIN_NAMESPACE
 
+// Note: this legacy ShaderEffect implementation is used only when running
+// directly with OpenGL. This is going to go away in the future (Qt 6?), since
+// the RHI path uses QQuickGenericShaderEffect always.
+
 namespace {
 
     enum VariableQualifier {
@@ -185,8 +189,8 @@ class MappedSlotObject: public QtPrivate::QSlotObjectBase
 public:
     typedef std::function<void()> PropChangedFunc;
 
-    explicit MappedSlotObject(PropChangedFunc func)
-        : QSlotObjectBase(&impl), _signalIndex(-1), func(func)
+    explicit MappedSlotObject(PropChangedFunc f)
+        : QSlotObjectBase(&impl), _signalIndex(-1), func(std::move(f))
     { ref(); }
 
     void setSignalIndex(int idx) { _signalIndex = idx; }
@@ -245,7 +249,11 @@ void QQuickOpenGLShaderEffectCommon::connectPropertySignals(QQuickItem *item,
                                                             const QMetaObject *itemMetaObject,
                                                             Key::ShaderType shaderType)
 {
-    QQmlPropertyCache *propCache = QQmlData::ensurePropertyCache(qmlEngine(item), item);
+    auto engine = qmlEngine(item);
+    if (!engine)
+        return;
+
+    QQmlPropertyCache *propCache = QQmlData::ensurePropertyCache(engine, item);
     for (int i = 0; i < uniformData[shaderType].size(); ++i) {
         if (signalMappers[shaderType].at(i) == 0)
             continue;
@@ -317,7 +325,8 @@ void QQuickOpenGLShaderEffectCommon::lookThroughShaderCode(QQuickItem *item,
                                                            Key::ShaderType shaderType,
                                                            const QByteArray &code)
 {
-    QQmlPropertyCache *propCache = QQmlData::ensurePropertyCache(qmlEngine(item), item);
+    auto engine = qmlEngine(item);
+    QQmlPropertyCache *propCache = (engine) ? QQmlData::ensurePropertyCache(engine, item) : nullptr;
     int index = 0;
     int typeIndex = -1;
     int typeLength = 0;
@@ -350,9 +359,11 @@ void QQuickOpenGLShaderEffectCommon::lookThroughShaderCode(QQuickItem *item,
             } else if (nameLength > srLen && qstrncmp("qt_SubRect_", s + nameIndex, srLen) == 0) {
                 d.specialType = UniformData::SubRect;
             } else {
-                if (QQmlPropertyData *pd = propCache->property(QString::fromUtf8(d.name), nullptr, nullptr)) {
-                    if (!pd->isFunction())
-                        d.propertyIndex = pd->coreIndex();
+                if (propCache) {
+                    if (QQmlPropertyData *pd = propCache->property(QString::fromUtf8(d.name), nullptr, nullptr)) {
+                        if (!pd->isFunction())
+                            d.propertyIndex = pd->coreIndex();
+                    }
                 }
                 const int mappedId = uniformData[shaderType].size() | (shaderType << 16);
                 mapper = new QtPrivate::MappedSlotObject([this, mappedId](){
@@ -703,8 +714,8 @@ void QQuickOpenGLShaderEffect::setBlending(bool enable)
 
 QVariant QQuickOpenGLShaderEffect::mesh() const
 {
-    return m_mesh ? qVariantFromValue(static_cast<QObject *>(m_mesh))
-                  : qVariantFromValue(m_meshResolution);
+    return m_mesh ? QVariant::fromValue(static_cast<QObject *>(m_mesh))
+                  : QVariant::fromValue(m_meshResolution);
 }
 
 void QQuickOpenGLShaderEffect::setMesh(const QVariant &mesh)
