@@ -3,7 +3,7 @@
 
 /*---
 esid: sec-array.prototype.reduceright
-es5id: 15.4.4.22-4-11
+es5id: 15.4.4.22-5-13
 description: >
     Array.prototype.reduceRight - the exception is not thrown if
     exception was thrown by step 3
@@ -26,5 +26,5 @@ Object.defineProperty(obj, "length", {
 });
 
 assert.throws(Test262Error, function() {
-  Array.prototype.reduceRight.call(obj, undefined);
+  Array.prototype.reduceRight.call(obj, function() {});
 });
