@@ -3,11 +3,11 @@
 
 /*---
 esid: sec-array.prototype.reduceright
-es5id: 15.4.4.22-8-b-iii-1-14
+es5id: 15.4.4.22-8-b-iii-1-13
 description: >
     Array.prototype.reduceRight - element to be retrieved is own
     accessor property that overrides an inherited accessor property on
-    an Array
+    an Array-like object
 ---*/
 
 var testResult = false;
@@ -18,22 +18,31 @@ function callbackfn(prevVal, curVal, idx, obj) {
   }
 }
 
-Object.defineProperty(Array.prototype, "2", {
+var proto = {
+  0: 0,
+  1: 1
+};
+
+Object.defineProperty(proto, "2", {
   get: function() {
     return 2;
   },
   configurable: true
 });
 
-var arr = [0, 1, , ];
+var Con = function() {};
+Con.prototype = proto;
 
-Object.defineProperty(arr, "2", {
+var child = new Con();
+child.length = 3;
+
+Object.defineProperty(child, "2", {
   get: function() {
     return "20";
   },
   configurable: true
 });
 
-arr.reduceRight(callbackfn);
+Array.prototype.reduceRight.call(child, callbackfn);
 
 assert(testResult, 'testResult !== true');
