@@ -2,10 +2,10 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 15.2.3.2-2-9
+es5id: 15.2.3.2-2-3
 description: >
     Object.getPrototypeOf returns the [[Prototype]] of its parameter
-    (Date)
+    (Object)
 ---*/
 
-assert.sameValue(Object.getPrototypeOf(Date), Function.prototype, 'Object.getPrototypeOf(Date)');
+assert.sameValue(Object.getPrototypeOf(Object), Function.prototype, 'Object.getPrototypeOf(Object)');
